@@ -1,13 +1,17 @@
 #!/usr/bin/env python3
 #
 #  testing.py
-"""
+r"""
 Functions for testing Sphinx extensions.
 
 .. extras-require:: testing
-	:__pkginfo__:
+	:pyproject:
 
 .. seealso:: Sphinx's own ``testing`` library: https://github.com/sphinx-doc/sphinx/tree/3.x/sphinx/testing
+
+.. latex:vspace:: 10px
+
+.. _pytest-regressions: https://pypi.org/project/pytest-regressions/
 """
 #
 #  Copyright © 2020-2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
@@ -60,28 +64,36 @@ Functions for testing Sphinx extensions.
 
 # stdlib
 import copy
+import re
+import sys
 import tempfile
+from functools import partial
 from types import SimpleNamespace
-from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple, Type, Union
+from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple, Type, Union, cast
 
 # 3rd party
 import pytest  # nodep
 import sphinx.application
-from bs4 import BeautifulSoup  # type: ignore
+from bs4 import BeautifulSoup  # type: ignore[import-untyped]
+from coincidence.regressions import (  # nodep
+		AdvancedFileRegressionFixture,
+		check_file_output,
+		check_file_regression
+		)
+from docutils import __version_info__ as docutils_version
 from docutils import nodes
 from docutils.parsers.rst import Directive, roles
 from docutils.transforms import Transform
 from domdf_python_tools.doctools import prettify_docstrings
 from domdf_python_tools.paths import PathPlus
 from domdf_python_tools.stringlist import StringList
-from domdf_python_tools.testing import check_file_output, check_file_regression
 from domdf_python_tools.typing import PathLike
-from pygments.lexer import Lexer  # type: ignore  # nodep
+from jinja2 import Template  # nodep
+from pygments.lexer import Lexer  # type: ignore[import-untyped]  # nodep
+from pytest_regressions.common import check_text_files  # nodep
 from pytest_regressions.file_regression import FileRegressionFixture  # nodep
 from sphinx.builders import Builder
-from sphinx.config import Config
 from sphinx.domains import Domain, Index
-# from sphinx.environment.collectors import EnvironmentCollector
 from sphinx.domains.python import PythonDomain
 from sphinx.environment.collectors import EnvironmentCollector
 from sphinx.events import EventListener
@@ -94,9 +106,21 @@ from sphinx.util import docutils
 from sphinx.util.typing import RoleFunction, TitleGetter
 
 # this package
-from sphinx_toolbox.utils import SphinxExtMetadata
+from sphinx_toolbox.utils import Config, SphinxExtMetadata
 
-__all__ = [
+# _ModuleWrapper unwrapping BS due to
+# https://github.com/sphinx-doc/sphinx/commit/8866adeacfb045c97302cc9c7e3b60dec5ca38fd
+
+try:
+	# 3rd party
+	from sphinx.deprecation import _ModuleWrapper  # type: ignore[attr-defined]
+	if isinstance(docutils, _ModuleWrapper):
+		docutils = docutils._module
+except ImportError:
+	# Unnecessary if unimportable
+	pass
+
+__all__ = (
 		"Sphinx",
 		"run_setup",
 		"RunSetupOutput",
@@ -106,7 +130,9 @@ __all__ = [
 		"check_asset_copy",
 		"HTMLRegressionFixture",
 		"html_regression",
-		]
+		"LaTeXRegressionFixture",
+		"latex_regression",
+		)
 
 
 class FakeBuilder(Builder):
@@ -144,7 +170,7 @@ class Sphinx:
 	def __init__(self):  # , buildername: str = "html"
 		self.registry = SphinxComponentRegistry()
 		self.config = Config({}, {})
-		self.events = EventManager(self)  # type: ignore
+		self.events = EventManager(self)  # type: ignore[arg-type]
 		self.html_themes: Dict[str, str] = {}
 		# self.builder = self.registry.create_builder(self, buildername)
 
@@ -198,6 +224,8 @@ class Sphinx:
 
 		The registered values are stored in the ``app.registry.translators`` dictionary.
 		(:class:`typing.Dict`\[:class:`str`\, :class:`typing.Type`\[:class:`docutils.nodes.NodeVisitor`\]]).
+
+		.. clearpage::
 		"""
 
 		self.registry.add_translator(name, translator_class, override=override)
@@ -228,7 +256,7 @@ class Sphinx:
 			self,
 			node: Type[nodes.Element],
 			figtype: str,
-			title_getter: TitleGetter = None,
+			title_getter: Optional[TitleGetter] = None,
 			override: bool = False,
 			**kwargs: Tuple[Callable, Callable],
 			) -> None:
@@ -236,7 +264,13 @@ class Sphinx:
 		Register a Docutils node class as a numfig target.
 		"""
 
-		self.registry.add_enumerable_node(node, figtype, title_getter, override=override)
+		# Sphinx's signature is wrong WRT Optional
+		self.registry.add_enumerable_node(
+				node,
+				figtype,
+				title_getter,
+				override=override,
+				)
 		self.add_node(node, override=override, **kwargs)
 
 	def add_directive(self, name: str, cls: Type[Directive], override: bool = False) -> None:
@@ -329,8 +363,8 @@ class Sphinx:
 			directivename: str,
 			rolename: str,
 			indextemplate: str = '',
-			parse_node: Callable = None,
-			ref_nodeclass: Type[nodes.TextElement] = None,
+			parse_node: Optional[Callable] = None,
+			ref_nodeclass: Optional[Type[nodes.TextElement]] = None,
 			objname: str = '',
 			doc_field_types: List = [],
 			override: bool = False,
@@ -339,6 +373,7 @@ class Sphinx:
 		Register a new object type.
 		"""
 
+		# Sphinx's signature is wrong WRT Optional
 		self.registry.add_object_type(
 				directivename,
 				rolename,
@@ -355,7 +390,7 @@ class Sphinx:
 			directivename: str,
 			rolename: str,
 			indextemplate: str = '',
-			ref_nodeclass: Type[nodes.TextElement] = None,
+			ref_nodeclass: Optional[Type[nodes.TextElement]] = None,
 			objname: str = '',
 			override: bool = False,
 			) -> None:
@@ -363,6 +398,7 @@ class Sphinx:
 		Register a new crossref object type.
 		"""
 
+		# Sphinx's signature is wrong WRT Optional
 		self.registry.add_crossref_type(
 				directivename,
 				rolename,
@@ -388,49 +424,41 @@ class Sphinx:
 
 	def add_js_file(self, filename: str, **kwargs: str) -> None:
 		"""
-		No-op for now.
+		Register a JavaScript file to include in the HTML output.
 
-		.. TODO:: Make this do something
+		.. versionadded:: 2.8.0
 		"""
 
-	#
-	# def add_js_file(self, filename: str, **kwargs: str) -> None:
-	# 	"""
-	# 	Register a JavaScript file to include in the HTML output.
-	# 	"""
-	#
-	# 	self.registry.add_js_file(filename, **kwargs)
+		self.registry.add_js_file(filename, **kwargs)
+
 	# 	if hasattr(self.builder, 'add_js_file'):
-	# 		self.builder.add_js_file(filename, **kwargs)  # type: ignore
+	# 		self.builder.add_js_file(filename, **kwargs)
 	#
 
 	def add_css_file(self, filename: str, **kwargs: str) -> None:
 		"""
-		No-op for now.
+		Register a stylesheet to include in the HTML output.
 
-		.. TODO:: Make this do something
+		.. versionadded:: 2.7.0
 		"""
 
-	# def add_css_file(self, filename: str, **kwargs: str) -> None:
-	# 	"""
-	# 	Register a stylesheet to include in the HTML output.
-	# 	"""
-	#
-	# 	self.registry.add_css_files(filename, **kwargs)
+		self.registry.add_css_files(filename, **kwargs)
+
 	# 	if hasattr(self.builder, 'add_css_file'):
-	# 		self.builder.add_css_file(filename, **kwargs)  # type: ignore
+	# 		self.builder.add_css_file(filename, **kwargs)
 
 	def add_latex_package(
 			self,
 			packagename: str,
-			options: str = None,
+			options: Optional[str] = None,
 			after_hyperref: bool = False,
 			) -> None:
 		"""
 		Register a package to include in the LaTeX source code.
 		"""
 
-		self.registry.add_latex_package(packagename, options, after_hyperref)  # type: ignore
+		# Sphinx's signature is wrong WRT Optional
+		self.registry.add_latex_package(packagename, cast(str, options), after_hyperref)
 
 	def add_lexer(self, alias: str, lexer: Type[Lexer]) -> None:
 		"""
@@ -506,7 +534,7 @@ class Sphinx:
 		Register a math renderer for HTML.
 		"""
 
-		self.registry.add_html_math_renderer(name, inline_renderers, block_renderers)  # type: ignore
+		self.registry.add_html_math_renderer(name, inline_renderers, block_renderers)
 
 	def setup_extension(self, extname: str) -> None:
 		"""
@@ -544,7 +572,7 @@ class RunSetupOutput(NamedTuple):
 	directives: Dict[str, Callable]  #: Mapping of directive names to directive functions.
 	roles: Dict[str, Callable]  #: Mapping of role names to role functions.
 	additional_nodes: Set[Type[Any]]  #: Set of custom docutils nodes registered in ``setup()``.
-	app: Sphinx  #: Instance of :class:`sphinx-toolbox.testing.Sphinx`.
+	app: Sphinx  #: Instance of :class:`sphinx_toolbox.testing.Sphinx`.
 
 
 _sphinx_dict_setup = Callable[[sphinx.application.Sphinx], Optional[Dict[str, Any]]]
@@ -554,6 +582,10 @@ _fake_metadata_setup = Callable[[Sphinx], Optional["SphinxExtMetadata"]]
 _setup_func_type = Union[_sphinx_dict_setup, _sphinx_metadata_setup, _fake_dict_setup, _fake_metadata_setup]
 
 
+class GenericNodeVisitor(nodes.NodeVisitor):
+	pass
+
+
 def run_setup(setup_func: _setup_func_type) -> RunSetupOutput:  # , buildername: str = "html"
 	"""
 	Function for running an extension's ``setup()`` function for testing.
@@ -567,17 +599,21 @@ def run_setup(setup_func: _setup_func_type) -> RunSetupOutput:  # , buildername:
 
 	app.add_domain(PythonDomain)
 
+	_additional_nodes = copy.copy(docutils.additional_nodes)
+	orig_gnv = nodes.GenericNodeVisitor
+
 	try:
-		_additional_nodes = copy.copy(docutils.additional_nodes)
+		nodes.GenericNodeVisitor = GenericNodeVisitor  # type: ignore[misc,assignment]
 		docutils.additional_nodes = set()
 
 		with docutils.docutils_namespace():
-			setup_ret = setup_func(app)  # type: ignore
-			directives = copy.copy(docutils.directives._directives)  # type: ignore
-			roles = copy.copy(docutils.roles._roles)  # type: ignore
+			setup_ret = setup_func(app)  # type: ignore[arg-type]
+			directives = copy.copy(docutils.directives._directives)  # type: ignore[attr-defined]
+			roles = copy.copy(docutils.roles._roles)  # type: ignore[attr-defined]
 			additional_nodes = copy.copy(docutils.additional_nodes)
 	finally:
 		docutils.additional_nodes = _additional_nodes
+		nodes.GenericNodeVisitor = orig_gnv  # type: ignore[misc]
 
 	return RunSetupOutput(setup_ret, directives, roles, additional_nodes, app)
 
@@ -618,10 +654,9 @@ def remove_html_link_tags(page: BeautifulSoup) -> BeautifulSoup:
 	return page
 
 
-def check_html_regression(page: BeautifulSoup, file_regression: FileRegressionFixture):
+def check_html_regression(page: BeautifulSoup, file_regression: FileRegressionFixture) -> None:
 	"""
-	Check an HTML page generated by Sphinx for regressions,
-	using `pytest-regressions <https://pypi.org/project/pytest-regressions/>`_
+	Check an HTML page generated by Sphinx for regressions, using `pytest-regressions`_.
 
 	:param page: The page to test.
 	:param file_regression: The file regression fixture.
@@ -633,7 +668,7 @@ def check_html_regression(page: BeautifulSoup, file_regression: FileRegressionFi
 		@pytest.mark.parametrize("page", ["index.html"], indirect=True)
 		def test_page(page: BeautifulSoup, file_regression: FileRegressionFixture):
 			check_html_regression(page, file_regression)
-	"""  # noqa D400
+	"""  # noqa: RST306
 
 	__tracebackhide__ = True
 
@@ -644,6 +679,9 @@ def check_html_regression(page: BeautifulSoup, file_regression: FileRegressionFi
 		if "_static/language_data.js" in str(div):
 			div.extract()
 
+	for div in page.select("div.sphinxsidebar"):
+		div.extract()
+
 	check_file_regression(
 			StringList(page.prettify()),
 			file_regression,
@@ -653,21 +691,42 @@ def check_html_regression(page: BeautifulSoup, file_regression: FileRegressionFi
 
 class HTMLRegressionFixture(FileRegressionFixture):
 	"""
-	Subclass of :class:`pytest_regressions.file_regression.FileRegressionFixture`
-	for checking HTML files.
+	Subclass of :class:`pytest_regressions.file_regression.FileRegressionFixture` for checking HTML files.
 
 	.. versionadded:: 2.0.0
-	"""  # noqa: D400
+	"""
 
-	def check(self, page: BeautifulSoup, **kwargs):  # type: ignore
+	def check(  # type: ignore[override]
+		self,
+		page: BeautifulSoup,
+		*,
+		extension: str = ".html",
+		jinja2: bool = False,
+		jinja2_namespace: Optional[Dict[str, Any]] = None,
+		**kwargs
+		) -> None:
 		r"""
-		Check an HTML page generated by Sphinx for regressions,
-		using `pytest-regressions <https://pypi.org/project/pytest-regressions/>`_
+		Check an HTML page generated by Sphinx for regressions, using `pytest-regressions`_.
 
 		:param page: The page to test.
+		:param jinja2: Whether to render the reference file as a jinja2 template.
+		:param jinja2_namespace: If ``jinja2`` is :py:obj:`True`,
+			a mapping of variable names to values to make available in the jinja2 template.
 		:param \*\*kwargs: Additional keyword arguments passed to
 			:meth:`pytest_regressions.file_regression.FileRegressionFixture.check`.
 
+		.. versionchanged:: 2.14.0  Added the ``jinja2`` keyword argument.
+		.. versionchanged:: 2.17.0  Added the ``jinja2_namespace`` keyword argument.
+
+		.. latex:clearpage::
+
+		When ``jinja2`` is :py:obj:`True`, the reference file will be rendered as a jinja2 template.
+		The template is passed the following variables:
+
+		* ``sphinx_version`` -- the Sphinx version number, as a tuple of integers.
+		* ``python_version`` -- the Python version number, in the form returned by :data:`sys.version_info`.
+		* ``docutils_version`` -- the docutils version number, as a tuple of integers (*New in version 2.16.0*).
+
 		**Example usage**
 
 		.. code-block:: python
@@ -675,7 +734,7 @@ class HTMLRegressionFixture(FileRegressionFixture):
 			@pytest.mark.parametrize("page", ["index.html"], indirect=True)
 			def test_page(page: BeautifulSoup, html_regression: HTMLRegressionFixture):
 				html_regression.check(page, file_regression)
-		"""  # noqa D400
+		"""  # noqa: RST306
 
 		__tracebackhide__ = True
 
@@ -686,20 +745,55 @@ class HTMLRegressionFixture(FileRegressionFixture):
 			if "_static/language_data.js" in str(div):
 				div.extract()
 
+		for div in page.select("div.sphinxsidebar"):
+			div.extract()
+
+		if sphinx.version_info >= (4, 3):  # pragma: no cover
+			for div in page.select("span.w"):
+				div.extract()
+			for div in page.select("span.p"):
+				if div.string == '=':
+					sibling = div.next_sibling
+					div.replace_with('')
+					sibling.replace_with(f"= {sibling.text}")
+
 		kwargs.pop("encoding", None)
 		kwargs.pop("extension", None)
 
+		if jinja2:
+
+			def check_fn(obtained_filename: PathPlus, expected_filename: PathPlus):  # noqa: MAN002
+				__tracebackhide__ = True
+
+				expected_filename = PathPlus(expected_filename)
+				template = Template(expected_filename.read_text())
+
+				expected_filename.write_text(
+						template.render(
+								sphinx_version=sphinx.version_info,
+								python_version=sys.version_info,
+								docutils_version=docutils_version,
+								**jinja2_namespace or {},
+								)
+						)
+
+				return check_text_files(obtained_filename, expected_filename, encoding="UTF-8")
+
+		else:
+			check_fn = partial(check_text_files, encoding="UTF-8")
+
 		super().check(
 				str(StringList(page.prettify())),
 				encoding="UTF-8",
-				extension=".html",
+				extension=extension,
+				check_fn=check_fn,
 				)
 
 
 @pytest.fixture()
-def html_regression(datadir, original_datadir, request) -> HTMLRegressionFixture:
+def html_regression(datadir, original_datadir, request) -> HTMLRegressionFixture:  # noqa: MAN001
 	"""
-	Returns a :class:`~.HTMLRegressionFixture` scoped to the test function.
+	Returns an :class:`~.HTMLRegressionFixture` scoped to the test function.
 
 	.. versionadded:: 2.0.0
 	"""
@@ -711,7 +805,7 @@ def check_asset_copy(
 		func: Callable[[sphinx.application.Sphinx, Exception], Any],
 		*asset_files: PathLike,
 		file_regression: FileRegressionFixture,
-		):
+		) -> None:
 	r"""
 	Helper to test functions which respond to Sphinx ``build-finished`` events and copy asset files.
 
@@ -729,11 +823,130 @@ def check_asset_copy(
 
 		fake_app = SimpleNamespace()
 		fake_app.builder = SimpleNamespace()
+		fake_app.builder.format = "html"
 		fake_app.outdir = fake_app.builder.outdir = tmp_pathplus
 
-		func(fake_app, None)  # type: ignore
+		func(fake_app, None)  # type: ignore[arg-type]
 
 		for filename in asset_files:
 			filename = tmp_pathplus / filename
 
 			check_file_output(filename, file_regression, extension=f"_{filename.stem}{filename.suffix}")
+
+
+_latex_date_re = re.compile(r"\\date{.*}")
+
+
+class LaTeXRegressionFixture(AdvancedFileRegressionFixture):
+	"""
+	Subclass of :class:`coincidence.regressions.AdvancedFileRegressionFixture` for checking LaTeX files.
+
+	.. versionadded:: 2.17.0
+	"""
+
+	def check(  # type: ignore[override]  # noqa: MAN002
+		self,
+		contents: Union[str, StringList],
+		*,
+		extension: str = ".html",
+		jinja2: bool = False,
+		jinja2_namespace: Optional[Dict[str, Any]] = None,
+		**kwargs
+		):
+		r"""
+		Check a LaTeX file generated by Sphinx for regressions,
+		using `pytest-regressions <https://pypi.org/project/pytest-regressions/>`__
+
+		:param contents:
+		:param jinja2: Whether to render the reference file as a jinja2 template.
+		:param jinja2_namespace: If ``jinja2`` is :py:obj:`True`,
+			a mapping of variable names to values to make available in the jinja2 template.
+		:param \*\*kwargs: Additional keyword arguments passed to
+			:meth:`pytest_regressions.file_regression.FileRegressionFixture.check`.
+
+		When ``jinja2`` is :py:obj:`True`, the reference file will be rendered as a jinja2 template.
+		The template is passed the following variables:
+
+		* ``sphinx_version`` -- the Sphinx version number, as a tuple of integers.
+		* ``python_version`` -- the Python version number, in the form returned by :data:`sys.version_info`.
+		* ``docutils_version`` -- the docutils version number, as a tuple of integers (*New in version 2.16.0*).
+
+		.. note::
+
+			Unlike standard HTML jinja2 templates,
+			this class expects the use of ``<`` and ``>`` rather than ``{`` and ``}``.
+
+			For example::
+
+				<% if foo %>
+				<# This should only happen on Tuesdays #>
+				<< foo.upper() >>
+				<% endif %>
+
+		**Example usage**
+
+		.. code-block:: python
+
+			@pytest.mark.sphinx("latex")
+			def test_latex_output(app: Sphinx, latex_regression: LaTeXRegressionFixture):
+				app.build()
+				output_file = app.outdir / "python.tex"
+				latex_regression.check(output_file.read_text())
+
+		"""  # noqa: D400
+
+		__tracebackhide__ = True
+
+		if jinja2:
+
+			def check_fn(obtained_filename: PathPlus, expected_filename: PathLike):  # noqa: MAN002
+				__tracebackhide__ = True
+
+				expected_filename = PathPlus(expected_filename)
+
+				template = Template(
+						expected_filename.read_text(),
+						block_start_string="<%",
+						block_end_string="%>",
+						variable_start_string="<<",
+						variable_end_string=">>",
+						comment_start_string="<#",
+						comment_end_string="#>",
+						)
+
+				expected_filename.write_text(
+						template.render(
+								sphinx_version=sphinx.version_info,
+								python_version=sys.version_info,
+								docutils_version=docutils_version,
+								**jinja2_namespace or {},
+								)
+						)
+
+				return check_text_files(obtained_filename, expected_filename, encoding="UTF-8")
+
+		else:
+			check_fn = partial(check_text_files, encoding="UTF-8")
+
+		new_contents = _latex_date_re.sub(
+				r"\\date{Mar 11, 2021}",
+				str(contents).replace("\\sphinxAtStartPar\n", ''),
+				)
+		new_contents = new_contents.replace("%% let collapsible ", "%% let collapsable ")  # changed in Sphinx 4.2
+
+		return super().check(
+				new_contents,
+				extension=".tex",
+				check_fn=check_fn,
+				)
+
+
+@pytest.fixture()
+def latex_regression(datadir, original_datadir, request) -> LaTeXRegressionFixture:  # noqa: MAN001
+	"""
+	Returns a :class:`~.LaTeXRegressionFixture` scoped to the test function.
+
+	.. versionadded:: 2.17.0
+	"""
+
+	return LaTeXRegressionFixture(datadir, original_datadir, request)
